"""Base class for materials."""

from typing import ClassVar, Optional

from gemd import MaterialTemplate, MaterialSpec, MaterialRun, PropertyAndConditions
from gemd.entity.util import make_instance
from gemd.enumeration import SampleType

from .base_element import BaseElement

# from openmsimodel.utilities.typing import ProcessDict, PropsAndCondsDict
from openmsimodel.utilities.attributes import finalize_template
from .process import Process

__all__ = ["Material"]


class Material(BaseElement):
    """
    Base element for materials.

    """

    _TempType = MaterialTemplate
    _SpecType = MaterialSpec
    _RunType = MaterialRun

    TEMPLATE: ClassVar[MaterialTemplate]

    def __init__(
        self,
        name: str,
        *,
        template: ClassVar[MaterialTemplate] = None,
        notes: Optional[str] = None,
        process: Optional[Process] = None,
        properties: Optional[list[PropertyAndConditions]] = None,
        sample_type: Optional[SampleType] = None
    ) -> None:
        super().__init__(name, template=template, notes=notes)

        self.set_process(process)

        if properties is None:
            properties = []

        self.update_properties_and_conditions(*properties, replace_all=True)

        self.set_sample_type(sample_type)

    @property
    def spec(self) -> MaterialSpec:
        """The underlying material spec."""
        return self._spec

    @property
    def run(self) -> MaterialRun:
        """The underlying material run."""
        return self._run

    @classmethod
    def from_spec_or_run(
        cls,
        name: str,
        *,
        notes: Optional[str] = None,
        spec: MaterialSpec = None,
        run: MaterialRun = None
    ) -> "Material":
        """
        Instantiate a `Material` from a spec or run with appropriate validation.

        Note that the spec's template will be set to the class template,
        and the run's spec will be set to this spec.

        if a
        """

        # if spec is None and run is None:
        #     raise ValueError("At least one of spec or run must be given.")

        # if spec is not None:
        #     template = spec.template
        # if run is not None:
        #     spec = run.spec
        #     template = spec.template
        # ensures
        if run is not None:
            if spec is None:
                spec = run.spec
            # template = spec.template
        else:
            if spec is None:
                raise ValueError("At least one of spec or run must be given.")
        template = spec.template

        material = cls(name, notes=notes, template=template)

        if spec is not None:
            if not isinstance(spec, MaterialSpec):
                raise TypeError("spec must be a MaterialSpec.")

            # template = spec.template

            material._spec = spec

            material.spec.name = name
            material.spec.notes = notes
            # material.spec.template = cls.TEMPLATE

            # material.update_properties_and_conditions()

        if run is not None:
            if not isinstance(run, MaterialRun):
                raise TypeError("run must be a MaterialRun.")

            material._run = run
            material.run.name = name
            material.run.notes = notes

            material.run.spec = material.spec

        else:
            material._run = make_instance(material.spec)

        return material

    def get_process_dict(self):
        """Get the names of the spec's and run's process."""

        return {"spec": self._spec.process.name, "run": self._run.process.name}

    def set_process(self, process: Optional[Process]) -> None:
        """
        Set the process that produces this material.

        process: {Process, None}
            Process instance whose spec and run will be set as the process for the material's
            spec and run, respectively. If ``None``, the material's spec and run process will be
            ``None``.
        """

        if process is not None:
            self._spec.process = process.spec
            self._run.process = process.run
        else:
            self._spec.process = None
            self._run.process = None

    def get_properties_and_conditions_dict(self):
        """
        Return a ``dict`` of material spec properties and conditions.
        The keys are the names of the properties.
        Each value is a ``dict`` with the keys ``'property'`` and ``'conditions'``.
        Each ``'property'`` key corresponds to another ``dict`` containing a value
        ``dict`` and origin ``str``.
        Each ``'condition'`` key corresponds to a ``dict`` in which the keys are
        the names of the conditions associated with a particular property and the
        values are value/origin ``dict`` s.
        """

        return self._prop_cond_dict(self._spec.properties)

    def update_properties_and_conditions(
        self, *properties: PropertyAndConditions, replace_all: bool = False
    ) -> None:
        """
        Change or add expected properties (with conditions) of the material spec.

        :param properties:
        :type properties: PropertyAndConditions
        :param replace_all:
            If ``True``, remove any existing properties before adding new ones.
        :type replace_all: bool
        :default replace_all: False
        """

        self._update_attributes(
            AttrType=PropertyAndConditions,
            attributes=properties,
            replace_all=replace_all,
            which="spec",
        )

    def remove_properties(self, *property_names: str) -> None:
        """
        Remove expected properties from the material spec by name.

        property_names: str
            The names of properties to remove.

        """

        self._remove_attributes(
            AttrType=PropertyAndConditions, attr_names=property_names, which="spec"
        )

    def get_sample_type(self) -> SampleType:
        """
        Get the sample type of the material run.
        sample_type: SampleType
        """

        return self._run.sample_type

    def set_sample_type(self, sample_type: Optional[SampleType]) -> None:
        """
        Set the sample type of the material run.

        sample_type: SampleType
        """

        self._run.sample_type = sample_type

    @staticmethod
    def _prop_cond_dict(
        spec_prop_conds: list[PropertyAndConditions],
    ):
        """Return a ``dict`` of material spec properties and conditions."""

        prop_cond_dict = {
            prop_cond.property.name: {
                "property": {
                    "value": prop_cond.property.value.as_dict(),
                    "origin": prop_cond.property.origin,
                },
                "conditions": {
                    cond.name: {"value": cond.value.as_dict(), "origin": cond.origin}
                    for cond in prop_cond.conditions
                },
            }
            for prop_cond in spec_prop_conds
        }

        return prop_cond_dict

    def to_form(self) -> str:
        pass



from typing import ClassVar

from gemd import (
    MaterialTemplate,
    ParameterTemplate,
    CategoricalBounds,
    NominalCategorical,
)

from openmsimodel.entity.base.material import Material
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
    _validate_temp_keys,
)

__all__ = ["InferredAlloyCompositions"]


class InferredAlloyCompositions(Material):
    """Class representing suggested compositions"""

    TEMPLATE: ClassVar[MaterialTemplate] = MaterialTemplate(
        "Inferred Alloy Compositions",
        description="Final selection of the bayesian optimziation algorithm, which can generate different number of compositions (i.e., 16 for VAM, 8 for DED) under various paramereters",
    )

    _ATTRS: ClassVar[AttrsDict] = _validate_temp_keys(TEMPLATE)

    # define_attribute(
    #     _ATTRS,
    #     template=ParameterTemplate(
    #         name='Supplier',
    #         bounds=CategoricalBounds(categories=[''])
    #     ),
    #     default_value=NominalCategorical('')
    # )

    finalize_template(_ATTRS, TEMPLATE)



'''General attribute utilities.'''

from typing import TypeAlias, Type, Union, Optional
import re
from datetime import datetime

from gemd import Condition, Parameter, Property, PerformedSource
from gemd.entity.template.attribute_template import AttributeTemplate
from gemd.entity.template.base_template import BaseTemplate

AttrTypes: TypeAlias = list[Union[Type[Condition], Type[Parameter], Type[Property]]]

def attr_template_dict(template: BaseTemplate) -> dict[str, AttributeTemplate]:
    '''
    Create a dictionary mapping attribute template names to the attributes themselves.
    
    Parameters
    ----------
    template: BaseTemplate
        An object template containing the attribute templates.

    Returns
    -------
    template_dict: dict
        Mapping of attribute template names to attribute templates.
    '''

    attrs = []

    try:
        attrs.extend(template.conditions)
    except AttributeError:
        pass
    try:
        attrs.extend(template.parameters)
    except AttributeError:
        pass
    try:
        attrs.extend(template.properties)
    except AttributeError:
        pass

    template_dict = {}

    for attr_temp, bounds in attrs:
        # deal with more restrictive bounds
        if bounds is not None:
            restricted_temp = attr_temp.__class__.build(attr_temp.dump())
            restricted_temp.bounds = bounds
            template_dict[attr_temp.name] = restricted_temp
        else:
            template_dict[attr_temp.name] = attr_temp

    return template_dict

def generate_tags(
    brand: Optional[str] = None,
    model: Optional[str] = None,
    instr_id: Optional[str] = None
    ) -> list[str]:
    '''Return a tags list containing one ``str`` from brand, model, and instrument id.'''

    tags = []

    if brand is not None:
        tag = str(brand)
        if model is not None:
            tag += f'::{model}'
            if instr_id is not None:
                tag += f'::{instr_id}'
        tags.append(tag)

    return tags

def generate_source(
    email: Optional[str] = None,
    iso_date: Optional[str] = None
    ) -> Optional[PerformedSource]:
    '''
    Generate a ``PerformedSource`` with a valid email address and an optional ISO date string.

    Parameters
    ----------
    email: str, optional
        A valid email address.
    date: str, optional
        A date string to be passed to ``datetime.fromisoformat``.

    Returns
    -------
    source: PerformedSource or None
        ``PerformedSource`` with email and date, or ``None`` if neither is supplied.

    Raises
    ------
    ValueError
        If `email` is invalid.
    '''

    if email is None and iso_date is None:
        return None

    if email is not None and not re.fullmatch(r'[^@]+@[^@]+\.[^@]+', email):
        raise ValueError(
            'Invalid email. Must contain a single "@" and at least one "." after the "@".'
        )

    if iso_date is not None:
        iso_date = datetime.fromisoformat(iso_date).isoformat(timespec='auto')

    return PerformedSource(email, iso_date)



import pymssql
import sqlalchemy as sqla
import pandas


class MSSQLDatabase:
    # wraps a Microsoft SQL Server database
    def __init__(self, AUTH, DATABASE=None):
        # AUTH should be a dict with some specific fields useful for a direct connection to the database
        self.AUTH = AUTH
        self.SERVER = AUTH["host"]
        if DATABASE is None:
            self.DATABASE = AUTH["database"]
        else:
            self.DATABASE = DATABASE
        self.ENGINE = self.__create_engine()

    # def executed_casjobs_query(self):
    #     pass

    def execPyMSSQL(self, statement):
        with pymssql.connect(
            self.SERVER, self.AUTH["user"], self.AUTH["pwd"], self.DATABASE
        ) as conn:
            cursor = conn.cursor()
            r = cursor.execute(statement)
            conn.commit()
        return r

    def execute_query(self, sql):
        if type(sql) == str:
            sql = sqla.text(sql)
        with self.ENGINE.connect() as conn:
            return pandas.read_sql(sql, conn)

    def execute_update(self, statement):
        if type(statement) == str:
            statement = sqla.text(statement)
        with self.ENGINE.connect() as conn:
            #         r=self.ENGINE.execute(statement)
            trans = conn.begin()
            try:
                result = conn.execute(statement)
                trans.commit()
            except:
                trans.rollback()
                raise
        return result

    def __create_engine(self):
        return sqla.create_engine(
            f"mssql+pymssql://{self.AUTH['user']}:{self.AUTH['pwd']}@{self.SERVER}:1433/{self.DATABASE}?charset=utf8"
        )

    def create_schema(self, schema):
        self.ENGINE.execute(sqla.schema.CreateSchema(schema))

    def get_source_id(self, ORG, schema="dbo"):
        sql = f"select source_id from {schema}.metadata_source where source_type='org' and organization_name='{ORG}'"
        with self.ENGINE.connect() as con:
            SOURCE_ID = pandas.read_sql(sql, con).source_id[0]
        return SOURCE_ID

    def drop_all_tables(self, schema, keep_tables=[]):
        tables = reversed(list(self.sorted_tables(schema).keys()))
        drop_all = "\n".join(
            [f"drop table {schema}.{t}" for t in tables if t not in keep_tables]
        )
        print(drop_all)

    #         execPyMSSQL(drop_all)

    def delete_from_all_tables(self, schema, keep_tables=[]):
        tables = reversed(list(self.sorted_tables(schema).keys()))
        delete_all = "\n".join(
            [f"delete from {schema}.{t}" for t in tables if t not in keep_tables]
        )
        print(delete_all)
        self.execPyMSSQL(delete_all)

    def sorted_tables(self, schema):
        # sort tables in a schema according to FK relationship
        # returns tables ordered such tha if table t1 refers to table t2, t2 will be earlier in the result than t1
        # the 'engine' variable should be result of a SQLAlchemy create_engine statement
        with self.ENGINE.connect() as conn:
            sql = f"""
              SELECT c.table_name, string_agg(c.column_name,',') within group(order by c.ordinal_position) as columns
              ,      sum(case when c.column_name='source_id' then 1 else 0 end) as has_source_id
              ,      max(case when COLUMNPROPERTY(object_id(c.TABLE_SCHEMA+'.'+c.TABLE_NAME), c.COLUMN_NAME, 'IsIdentity') = 1 then c.column_name else null end) as identity_column
              from information_Schema.tables t
              join information_schema.columns c on c.table_schema=t.table_schema and c.table_name=t.table_name
             where t.table_schema = '{schema}' and t.table_type='BASE TABLE'
             group by c.table_name
            """
            tables = pandas.read_sql(sql, conn)
            tables = {
                t.table_name: {
                    "columns": t.columns,
                    "identity_column": t.identity_column,
                    "has_source_id": t.has_source_id,
                    "FKs": set(),
                }
                for t in tables.itertuples()
            }

            sql = f"""
            SELECT OBJECT_NAME(fk.parent_object_id) as from_table
            ,      OBJECT_NAME(fk.referenced_object_id) to_table
              FROM sys.foreign_keys fk
              INNER JOIN sys.foreign_key_columns fkc 
                 ON fkc.constraint_object_id = fk.object_id
              INNER JOIN sys.columns c1 
                 ON fkc.parent_column_id = c1.column_id 
                AND fkc.parent_object_id = c1.object_id
              INNER JOIN sys.columns c2 
                 ON fkc.referenced_column_id = c2.column_id 
                AND fkc.referenced_object_id = c2.object_id
              inner join sys.schemas s on s.schema_id=fk.schema_id
          where s.name='{schema}'
            """
            FKs = pandas.read_sql(sql, conn)
            for f in FKs.itertuples():
                tables[f.from_table]["FKs"].add(f.to_table)

        ordered = []
        done = set()

        def topsort(t):
            if t not in tables:
                return
            refs = tables[t]["FKs"]
            for r in refs:
                if r in done:
                    continue
                topsort(r)
            ordered.append(t)
            done.add(t)

        for t in tables:
            if t not in done:
                topsort(t)
        return {
            t: {
                "columns": tables[t]["columns"],
                "identity_column": tables[t]["identity_column"],
                "has_source_id": True if tables[t]["has_source_id"] == 1 else False,
                "FKs": tables[t]["FKs"] if len(tables[t]["FKs"]) > 0 else None,
            }
            for t in ordered
        }



# from .alloy import Alloy
# from .composition import Composition
# from .inferred_alloy_compositions import InferredAlloyCompositions
# from .summary_sheet import SummarySheet
# from .traveler_sample import TravelerSample
# from .traveler import Traveler

# __all__ = [
#     "Alloy",
#     "Composition",
#     "InferredAlloyCompositions",
#     "SummarySheet",
#     "TravelerSample",
#     "Traveler",
# ]

# del alloy
# del composition
# del inferred_alloy_compositions
# del summary_sheet
# del traveler_sample
# del traveler



from pathlib import Path
import os


def is_not_hidden(path):
    return not path.name.startswith(".")


class FolderOrFile:
    """
    class that represents a folder and its elements, or a file.
    It uses recursion to explore and record folder hiearchies that can be exploited
    for modelling. For example, if a team follows a protocol for data storage, naming, etc, using
    folder structures.
    """

    display_filename_prefix_middle = "├──"
    display_filename_prefix_last = "└──"
    display_parent_prefix_middle = "    "
    display_parent_prefix_last = "│   "

    # def __init__(self, *args, **kwargs):
    def __init__(self, root, parent_path, is_last):
        """
        this class represents a folder or a file, and is used in a recursive approach to
        build trees, navigate them or display them swiftly.
        :param path: the path to the folder/file to display and build tree from
        :param parent_path: path to the parent folder of current object
        :param is_last: is this the last element in the tree

        """
        self.root = Path(str(root))
        self.parent = parent_path
        self.is_last = is_last
        if self.parent:
            self.depth = self.parent.depth + 1
        else:
            self.depth = 0

    @property
    def displayname(self):
        """
        displays the name of the folder or file
        """
        if self.root.is_dir():
            return self.root.name + "/"
        return self.root.name

    @classmethod
    def make_tree(self, cls, root, parent=None, is_last=False, criteria=None):  # FIXME
        """
        powerful function that builds trees recursively.
        It passes the current item, displayable_root, as the 'parent' parameters to its children to navigate down the tree.
        It stops the recursion when there are no more children, or is_last=False.
        It can use the criteria to filter the children
        :param root: the path to the folder/file to display and build tree from
        :param parent: parent object of current item
        :param is_last: is this the last element in the tree
        :criteria: a function that takes a path and returns True if the path should be included in the tree
        """
        root = Path(str(root))
        criteria = criteria or cls._default_criteria

        displayable_root = cls(root, parent, is_last)
        yield displayable_root

        # if directory, list of elements in lower case
        children = sorted(
            list(path for path in root.iterdir() if criteria(path)),
            key=lambda s: str(s).lower(),
        )

        count = 1  # used to track nb of elements, not DEPTH
        for path in children:
            is_last = count == len(children)
            if path.is_dir():  # directory vs file
                yield from cls.make_tree(
                    cls=cls,
                    root=path,
                    parent=displayable_root,
                    is_last=is_last,
                    criteria=criteria,
                )
            else:
                yield cls(path, displayable_root, is_last)
            count += 1

    @classmethod
    def _default_criteria(cls, path):
        return True

    def displayable(self):
        """
        functions that returns a nicely formatted string that display the tree items and hierarchy.
        It navigates up to the tree in a loop by reading the parent attribute one at a time, and
        extracting its displayname.
        Finally, it renders all the display names (which are reversed to be in the correct order).
        """
        if self.parent is None:
            return self.displayname

        # distinguishes the printing of tree symbols from the last one and the rests
        _filename_prefix = (
            self.display_filename_prefix_last
            if self.is_last
            else self.display_filename_prefix_middle
        )

        # attaches correct tree symbols with actual file/folder names
        parts = ["{!s} {!s}".format(_filename_prefix, self.displayname)]

        parent = self.parent
        while (
            parent and parent.parent is not None
        ):  # TODO: update because empty folders are considered as None
            parts.append(
                self.display_parent_prefix_middle
                if parent.is_last
                else self.display_parent_prefix_last
            )
            parent = parent.parent

        # because children are appended first and parents last, a reversal is necessary
        return "".join(reversed(parts))



"""Base class for ingredient."""

from abc import ABC
from typing import ClassVar, Optional

from openmsimodel.utilities.typing import Spec, Run

# from openmsimodel.utilities.typing import ProcessDict, PropsAndCondsDict

from gemd import IngredientSpec, IngredientRun, PropertyAndConditions
from gemd.entity.util import make_instance
from gemd.enumeration import SampleType

from .base_element import BaseElement
from .process import Process

__all__ = ["Ingredient"]


class Ingredient(BaseElement):
    """
    Base element for ingredients.

    """

    _SpecType = IngredientSpec
    _RunType = IngredientRun

    def __init__(self, name: str, *, notes: Optional[str] = None) -> None:
        # BaseElement.__init__(self, name)
        super(ABC, self).__init__()
        self.name = name
        self._spec: Spec = self._SpecType(name=name, notes=notes)
        self._run: Run = make_instance(self._spec)

    @property
    def spec(self) -> IngredientSpec:
        """The underlying process spec."""
        return self._spec

    @property
    def run(self) -> IngredientRun:
        """The underlying process run."""
        return self._run

    @classmethod
    def from_spec_or_run(
        cls,
        name: str,
        *,
        notes: Optional[str] = None,
        spec: IngredientSpec = None,
        run: IngredientRun = None
    ) -> "Process":
        """
        Instantiate a `Process` from a spec or run with appropriate validation.

        Note that the spec's template will be set to the class template,
        and the run's spec will be set to this spec.
        """
        pass

    def to_form(self) -> str:
        pass

    # TODO: from_material
    # TODO: from_material_run_or_spec



from typing import ClassVar

from gemd import ProcessTemplate

from openmsimodel.entity.base.process import Process
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
)

__all__ = ["SettingTravelerSample"]


class SettingTravelerSample(Process):
    """Class representing singular sample extraction from sample"""

    TEMPLATE: ClassVar[ProcessTemplate] = ProcessTemplate(
        name="Setting Traveler Sample",
        description="""extracting a single sample, such as T01, from the traveler
                """,
    )

    _ATTRS: ClassVar[AttrsDict] = {"conditions": {}, "parameters": {}}

    finalize_template(_ATTRS, TEMPLATE)



from .measure_dimensions import MeasureDimensions
from .mounting_and_polishing import MountingAndPolishing
from .ni import NI
from .sem import SEM
from .tensile import Tensile
from .weighting import Weighting
from .xrd import XRD

__all__ = [
    "MeasureDimensions",
    "MountingAndPolishing",
    "NI",
    "SEM",
    "Tensile",
    "Weighting",
    "XRD",
]

del measure_dimensions
del mounting_and_polishing
del ni
del sem
del tensile
del weighting
del xrd



from yfiles_jupyter_graphs import GraphWidget
import networkx as nx
from webcolors import name_to_hex


def launch_graph_widget(g):
    if type(g) == str:
        if g.endswith(".dot"):
            dot = nx.nx_pydot.read_dot(g)
            g = nx.Graph(dot)
        elif g.endswith(".graphml"):
            g = nx.read_graphml(g)
    elif g.__class__ is not None and g.__class__.__name__ == "AGraph":
        g = nx.nx_agraph.from_agraph(g)

    def node_color_mapping(index, node):
        try:
            color = node["properties"]["color"]
        except:
            color = "white"
        return name_to_hex(color)

    def edge_color_mapping(index, node):
        try:
            color = node["properties"]["color"]
        except:
            color = "white"
        return name_to_hex(color)

    w = GraphWidget(graph=g)
    w.set_node_color_mapping(node_color_mapping)
    # w.set_edge_color_mapping(node_color_mapping)
    w.directed = True
    w.hierarchic_layout()
    w.show()



# from .add_material import AddMaterial
# from .aggregate_summary_sheet import AggregateSummarySheet
# from .arc_melting import ArcMelting
# from .forging import Forging
# from .homogenization import Homogenization
# from .infer_compositions import InferCompositions
# from .mixing import Mixing
# from .select_composition import SelectComposition
# from .setting_traveler_sample import SettingTravelerSample
# from .setting_traveler import SettingTraveler

# __all__ = [
#     "AddMaterial",
#     "AggregateSummarySheet",
#     "ArcMelting",
#     "Forging",
#     "Homogenization",
#     "InferCompositions",
#     "Mixing",
#     "SelectComposition",
#     "SettingTravelerSample",
#     "SettingTraveler",
# ]

# del add_material
# del aggregate_summary_sheet
# del arc_melting
# del forging
# del homogenization
# del infer_compositions
# del mixing
# del select_composition
# del setting_traveler_sample
# del setting_traveler



import setuptools

version = "0.1.0"

setupkwargs = dict(
    name="openmsimodel",
    version=version,
    description="OpenMSIModel uses the GEMD (Graphical Expression of Material Data) format to interact with generalized laboratory, analysis, and computational materials data. It allows to structure real scientific workflows into meaningful data structures, model them in graph and relational databases, explore on interactive graphical interfaces, and build long-term, shareable assets stores.",
    url="https://github.com/openmsi/openmsimodel",
    author="Ali Rachidi",
    author_email="arachid1@jhu.edu",
    license="MIT",
    packages=setuptools.find_packages(include=["openmsimodel*"]),
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "open_db=openmsimodel.db.open_db:main",
            "open_graph=openmsimodel.graph.open_graph:main",
        ]
    },
    python_requires=">=3.7,<3.10",
    install_requires=[
        "gemd==1.13.0",
        "networkx",
        "pymssql==2.2.8",
        "SQLAlchemy==2.0.17",
        "PyInquirer",
        "prompt_toolkit",
        "numpy",
        "pandas",
        "methodtools",
        "matplotlib",
        "yfiles_jupyter_graphs",
        "ipython",
        "ipykernel",
        "openpyxl",
        "graphviz",
        "questionary",
        "webcolors",
    ],
    extras_require={},
)

setupkwargs["extras_require"]["all"] = sum(setupkwargs["extras_require"].values(), [])

setuptools.setup(**setupkwargs)



'''Base for special attribute classes.'''

from typing import ClassVar, Union
from abc import ABC, abstractmethod
import re

import numpy as np
import pandas as pd

from gemd.entity.template.attribute_template import AttributeTemplate
from gemd.entity.attribute.base_attribute import BaseAttribute
from gemd.entity.object.base_object import BaseObject

from .utils import AttrTypes

__all__ = ['BaseSpecialAttrs']

class BaseSpecialAttrs(ABC):
    '''
    Abstract class for classes that provide convenient interfaces to certain types
    of attributes attached to GEMD objects.
    '''

    RE: ClassVar[re.Pattern]

    @abstractmethod
    def concat(self, *args: pd.DataFrame) -> None:
        '''
        Concatenate new data.
        
        Indices in the new data are respected and will replace rows with
        identical indices in the existing data.
        '''

    @abstractmethod
    def remove(self, index: int) -> None:
        '''Remove a row from the attributes by index, then reindex.'''

    @abstractmethod
    def update_object(self) -> None:
        '''
        Update the ``<Object>Spec`` or ``<Object>Run`` used to initialize
        with the attributes.

        This is performed in-place on the existing <object> and will
        overwrite any attributes added after ``<SubClass>`` initialization.
        '''

    @classmethod
    @abstractmethod
    def object_from_file(cls, *args, **kwargs) -> tuple[BaseObject, BaseObject]:
        '''
        Create an empty ``<Object>Spec`` and filled ``<Object>Run`` from a delimited file.
        
        Add parameters to the method definition as needed, e.g. filename, url,
        template, etc.
        '''

    @classmethod
    @abstractmethod
    def _parse(cls, attrs: list[BaseAttribute]):
        '''
        Match each attribute's name with regular expression.

        Should return ``DataFrame``(s) as appropriate and a list of other
        ``BaseAttribute``s.
        '''

    @classmethod
    @abstractmethod
    def _add_attr_to_df(cls, *args, **kwargs) -> None:
        '''Given a matching attribute, add it to a ``DataFrame`` as appropriate.'''

    @classmethod
    @abstractmethod
    def _reindex_df(cls, *args: pd.DataFrame):
        '''Add in any missing integer indices and sort.'''

    @classmethod
    @abstractmethod
    def _reindex_all_dfs(self) -> None:
        '''Ensure that each ``DataFrame`` has the same (maximum) number of indices.'''

    @abstractmethod
    def _reindex_all_attrs(self) -> None:
        '''Rename all attributes according to where they are in their ``DataFrame``.'''

    @staticmethod
    def _extract_attrs_from_df(df: pd.DataFrame) -> list[BaseAttribute]:
        '''Return a list of attributes from a ``DataFrame``.'''

        return [attr for col in df for attr in col if not pd.isna(attr)]

    @classmethod
    @abstractmethod
    def _create_attrs_from_df(
        cls,
        df: pd.DataFrame,
        template_dict: dict[str, AttributeTemplate],
        attr_types: AttrTypes,
        columns: list[str],
        names: list[str],
        units: list[Union[str, None]],
        origins: list[str]
        ):
        '''Create attributes from the `df` values.'''



from typing import ClassVar

from gemd import MeasurementTemplate, PropertyTemplate, RealBounds, NominalReal

from openmsimodel.entity.base.measurement import Measurement
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
)

__all__ = ["MeasureDimensions"]


class MeasureDimensions(Measurement):
    """Class representing measurement of diemsniosn as a measurement"""

    TEMPLATE: ClassVar[MeasurementTemplate] = MeasurementTemplate(
        name="MeasureDimensions",
        description="The measurement of dimension of an alloy in its form, often ingot",
    )

    _ATTRS: ClassVar[AttrsDict] = {"conditions": {}, "parameters": {}, "properties": {}}

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Length",
            bounds=RealBounds(0, 100, "cm"),
            description="Length of alloy",
        ),
    )

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Width",
            bounds=RealBounds(0, 50, "cm"),
            description="Width of alloy",
        ),
    )

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Thickness",
            bounds=RealBounds(0, 25, "cm"),
            description="Thickness of alloy",
        ),
    )

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Thickness Reduction",
            bounds=RealBounds(-150, 150, "cm"),
            description="Thickness of alloy",
        ),
    )

    finalize_template(_ATTRS, TEMPLATE)



"""
Class defining the general OpenMSIStream workflow for running from the command line
(or as a Service/daemon)
"""

# imports
from abc import ABC, abstractmethod
from openmsimodel.utilities.has_arguments import HasArguments
from openmsimodel.utilities.has_argument_parser import HasArgumentParser


class Runnable(HasArguments, HasArgumentParser, ABC):
    """
    Abstract base class for any child classes that want to define some behavior
    for running from the command line (i.e. as a module)
    """

    ARGUMENT_PARSER_TYPE = None

    @classmethod
    def get_argument_parser(cls, *args, **kwargs):
        """
        Get the argument parser used to run the code

        :param args: Any arguments to this method are names of arguments recognized
            by Argument parsers of the :attr:`~Runnable.ARGUMENT_PARSER_TYPE` type
        :type args: list
        :param kwargs: Any keyword arguments to this method define custom default values
            for their given arguments, whose names must be recognized by Argument parsers
            of the :attr:`~Runnable.ARGUMENT_PARSER_TYPE` type
        :type kwargs: dict

        :return: An argument parser of the :attr:`~Runnable.ARGUMENT_PARSER_TYPE` type to use
            for the object
        """
        parser = cls.ARGUMENT_PARSER_TYPE(*args, **kwargs)
        cl_args, cl_kwargs = cls.get_command_line_arguments()
        parser.add_arguments(*cl_args, **cl_kwargs)
        return parser

    @classmethod
    @abstractmethod
    def run_from_command_line(cls, args=None):
        """
        Child classes should implement this function to do whatever it is they do
        when they run from the command line

        :param args: the list of arguments to send to the parser instead of
            getting them from sys.argv
        :type args: list
        """
        raise NotImplementedError
